#include "Base/Allocator.h"

namespace lstore {

Allocator::Allocator(usize block_size, usize block_count, bool infinity_pool)
    : block_size_(block_size), block_count_(block_count),
      is_infinity_pool_(infinity_pool) {
  current_pool_ = new uint8[block_size_ * block_count_];
  if (is_infinity_pool_) {
    pools_.push_back(current_pool_);
  }
}

Allocator::~Allocator() {
  if (is_infinity_pool_) {
    for (auto* pool : pools_) {
      delete[] pool;
    }
  } else {
    delete[] current_pool_;
  }
}

void* Allocator::allocate() {
  spin_.lock();
  void* free_block = pop();
  if (!free_block) { // If can't obtain existing block then get a new one
    if (pool_index_ < block_count_) {
      uint8* mem = current_pool_ + (pool_index_ * block_size_);
      pool_index_++;
      free_block = reinterpret_cast<void*>(mem);
    } else {
      if (is_infinity_pool_) { // allocate new pool
        current_pool_ = new uint8[block_size_ * block_count_];
        pools_.push_back(current_pool_);
        pool_index_ = 1;
        free_block = reinterpret_cast<void*>(current_pool_);
      }
    }
  }
  spin_.unlock();
  return free_block;
}

void Allocator::deallocate(void* block) {
  if (block == nullptr) {
    return;
  }
  spin_.lock();
  push(block);
  spin_.unlock();
}

void Allocator::push(void* ptr) {
  Block* block = reinterpret_cast<Block*>(ptr);
  block->next = free_blocks_head_;
  free_blocks_head_ = block;
}

void* Allocator::pop() {
  Block* block = nullptr;
  if (free_blocks_head_) {
    block = free_blocks_head_;
    free_blocks_head_ = free_blocks_head_->next;
  }
  void* ptr = reinterpret_cast<void*>(block);
  return ptr;
}

} // namespace lstore
