#include "Catalog/Catalog.h"

#include <filesystem>
#include <fstream>
#include <iomanip>
#include <memory>

#include "Base/json_utils.h"

namespace lstore {

constexpr std::string kCatalogFileName = "catalog.json";

Catalog::Catalog() {
  readFromFile();
}

Catalog::~Catalog() {
  writeToFile();
}

TablePtr& Catalog::getTable(oid id) {
  return tables_[id];
}

bool Catalog::containsTable(oid id) {
  return tables_.contains(id);
}

FilePtr& Catalog::getFile(oid id) {
  return files_[id];
}

TablePtr& Catalog::createTable(const ColumnDescs& user_columns, const std::string& name) {
  oid table_id = table_id_seq_++;
  oid first_fil_id = file_id_seq_++;
  TablePtr table = std::make_shared<Table>(user_columns, name, table_id, first_fil_id);
  tables_[table_id] = table;
  const ColumnDescs& columns = table->getColumns();
  for (usize i = 0; i < columns.size(); i += 1) {
    auto baseFileName = columns[i].getName() + "_b.data";
    oid base_file_id = table->getColumnBaseFileOid(static_cast<uint32>(i));
    FilePtr baseFile = std::make_shared<File>(baseFileName, base_file_id);
    files_[base_file_id] = baseFile;

    auto tailFileName = columns[i].getName() + "_t.data";
    oid tail_file_id = base_file_id + 1;
    FilePtr tailFile = std::make_shared<File>(tailFileName, tail_file_id);
    files_[tail_file_id] = tailFile;
  }
  namespace fs = std::filesystem;
  fs::path p = table->getName();
  fs::create_directory(p);
  return tables_[table_id];
}

template <typename T>
static void map_from_json(const json& j, std::unordered_map<oid, std::shared_ptr<T>>& v) {
  for (auto& el : j) {
    auto t = std::make_shared<T>();
    from_json(el, *t);
    v[t->getId()] = t;
  }
}

template<typename T>
static void map_to_json(json& j, const std::unordered_map<oid, std::shared_ptr<T>>& v) {
  for (auto& [id, val] : v) {
    json t;
    to_json(t, *val);
    j.push_back(t);
  }
}

void Catalog::readFromFile() {
  std::ifstream fin(kCatalogFileName);
  if (!fin) {
    return;
  }
  json data = json::parse(fin);

  table_id_seq_ = data.at("tableIdSeq");
  file_id_seq_ = data.at("fileIdSeq");

  map_from_json(data.at("tables"), tables_);
  map_from_json(data.at("files"), files_);
}

void Catalog::writeToFile() {
  json j;
  j["tableIdSeq"] = table_id_seq_;
  j["fileIdSeq"] = file_id_seq_;

  j["tables"] = json::array();
  map_to_json(j["tables"], tables_);
  for (auto& [oid, table] : tables_) {
    table->init();
  }

  j["files"] = json::array();
  map_to_json(j["files"], files_);

  std::ofstream fout(kCatalogFileName);
  fout << std::setw(4) << j << std::endl;
}

} // namespace lstore
