#include "Catalog/Table.h"

#include "DataTypes/TypeRegistry.h"
#include "Indexes/BTreeIndex.h"

namespace lstore {

Table::Table(const ColumnDescs& user_columns, const std::string& name, oid id,
            oid first_file_oid)
    : name_(name), id_(id), first_file_oid_(first_file_oid) {
  fillByUserColumns(user_columns);
  init_index();
}

void Table::init() {
  ColumnDescs user_columns = std::move(columns_);
  fillByUserColumns(user_columns);
  init_index();
}

void from_json(const json& j, Table& t) {
  j.at("name").get_to(t.name_);
  j.at("id").get_to(t.id_);
  j.at("first_file_oid").get_to(t.first_file_oid_);
  vector_from_json(j.at("columns"), t.columns_);
}

void to_json(json& j, Table& t) {
  j["name"] = t.name_;
  j["id"] = t.id_;
  j["first_file_oid"] = t.first_file_oid_;
  j["columns"] = json::array();
  vector_to_json(j["columns"], t.columns_);
}

oid Table::getColumnBaseFileOid(uint32 column_index) const {
  return first_file_oid_ + column_index * 2;
}

void Table::fillByUserColumns(const ColumnDescs& user_columns) {
  using namespace std::string_literals;
  usize max_columns_count = kMetaColumnsCount + user_columns.size() * 2;
  columns_.reserve(max_columns_count);
  columns_.push_back(ColumnDesc{"Indirection_meta"s, TypeId::UInt64});
  columns_.push_back(ColumnDesc{"Schema_encoding_meta"s, TypeId::UInt64});
  columns_.push_back(ColumnDesc{"Start_time_meta"s, TypeId::UInt64});
  columns_.push_back(ColumnDesc{"Page_type_dependent_meta"s, TypeId::UInt64});
  user_column_indexes_.reserve(user_columns.size());
  usize user_column_index = kMetaColumnsCount;
  for (auto column : user_columns) {
    user_column_indexes_.push_back(user_column_index);
    DataTypePtr type = TypeRegistry::getType(column.getType());
    if (!type->haveMaximumSizeOfValue()) {
      std::string meta_column_name = column.getName() + "_meta"s;
      columns_.push_back(ColumnDesc{meta_column_name, TypeId::UInt64}); // TODO: change type
      user_column_index++;
    }
    columns_.push_back(column);
    user_column_index++;
  }
}

void Table::init_index() {
  std::string file_name = name_ + ".index";
  index_ = std::make_shared<BTreeIndex>(id_, 0, file_name);
}

} // namespace lstore
