#ifndef LSTORE_CORE_VALUE_H
#define LSTORE_CORE_VALUE_H

#include <string>
#include <type_traits>
#include <utility>
#include <variant>
#include <vector>

#include "Base/types.h"
#include "DataTypes/DataTypes.h"
#include "DataTypes/TypeId.h"

namespace lstore {

class Value;

template <typename T>
concept not_value_or_stringlike
    = (!std::is_same_v<std::decay_t<T>, Value>
       && !std::is_same_v<std::decay_t<T>, std::string>
       && !std::is_same_v<std::decay_t<T>, std::string_view>
       && !std::is_same_v<std::decay_t<T>, const char*>);

/**
 * Union of several types.
 * Used to represent a single value of one of several types in memory.
 */
class Value {
public:
  template <typename T>
  requires not_value_or_stringlike<T>
  Value(T&& v);

  Value(const Value& v) { assign(v); }
  Value(Value&& v) { assign(std::move(v)); }
  Value& operator=(const Value& v);
  Value& operator=(Value&& v);

  /**
   * Create a string inplace.
   */
  Value(std::string_view str) { assignString(str.data(), str.size()); }
  Value(const std::string& str) { Value(std::string_view{str}); }
  Value(std::string&& str) { assignString(std::move(str)); }
  Value(const char* str) { Value(std::string_view{str}); }

  /**
   * Allows expressions like
   * Value v = 1;
   */
  template <typename T>
  requires not_value_or_stringlike<T>
  Value& operator=(T&& v);

  Value& operator=(std::string_view str);
  Value& operator=(const std::string& str) { return *this = std::string_view{str}; }
  Value& operator=(std::string&& str);
  Value& operator=(const char* str) { return *this = std::string_view{str}; }

  ~Value() { destroy(); }

  TypeId getType() const { return type_; }

  template <typename T>
  T& get() { return std::get<T>(storage_); }

  template <typename T>
  const T& get() const { return std::get<T>(storage_); }

private:
  std::variant<UInt8, UInt16, UInt32, UInt64, Int8, Int16, Int32, Int64,
    Float32, Float64, String
    > storage_;

  TypeId type_;

  void assignString(std::string&& str);
  void assignString(const char* data, usize size);

  void assign(const Value& x);
  void assign(Value&& x);

  void destroy();
};

using Row = std::vector<Value>;

template <typename T>
requires not_value_or_stringlike<T>
Value::Value(T&& v) {
  storage_ = v;
  type_ = static_cast<TypeId>(storage_.index() + 1);
  //type_ = TypeToTypeId<T>;
}

template <typename T>
requires not_value_or_stringlike<T>
Value& Value::operator=(T&& v) {
  storage_ = std::move(v);
  type_ = static_cast<TypeId>(storage_.index() + 1);
  //type_ = TypeToTypeId<T>;
  return *this;
}

} // namespace lstore

#endif // LSTORE_CORE_VALUE_H
