#include "Pages/InsertRange.h"

#include "DataTypes/TypeRegistry.h"
#include "Pages/ValueLoc.h"

#include <iostream>

namespace lstore {

InsertRange::InsertRange(TablePtr& table, rid first)
    : first_(first), current_(first), table_(table) {
  usize column_count = table_->getColumns().size();
  ranges_.push_back(createPageRange(column_count));
}

PageRange InsertRange::createPageRange(usize column_count) {
  PageRange range(column_count);
  for (auto& page_vector: range) {
    page_vector.push_back(std::move(Page{}));
  }
  return range;
}

void InsertRange::insert(xid id, const Row& row, MergeManager& merge_manager) {
  (void) merge_manager; // TODO
  UInt64 key = row[0].get<UInt64>();
  IndexPtr& index = table_->getIndex();
  spin_.lock();
  rid cur = current_;
  current_ += 1;
  rid row_id = cur - first_;
  usize cur_range_index = row_id / kBaseRangeSize;
  if (ranges_.size() == cur_range_index) {
    ranges_.push_back(createPageRange(ranges_[0].size()));
  }
  index->insert(key, cur);
  spin_.unlock();

  InsertArgs args = {
    .row_id = row_id,
    .cur_range = ranges_[cur_range_index],
    .id = id,
  };
  insertMetaData(args);
  insertUserData(args, row);
}

bool InsertRange::isFull() {
  spin_.lock();
  rid diff = current_ - first_;
  spin_.unlock();
  return diff == kInsertRangeSize;
}

void InsertRange::insertMetaData(InsertArgs& args) {
  UInt64 zero = 0;
  Value zero_val = zero;

  DataTypePtr& meta_columns_type = TypeRegistry::getType(TypeId::UInt64);
  usize type_size = meta_columns_type->getSizeOfValueInMemory();
  ValueLoc loc{args.row_id, type_size};

  PageRange& range = args.cur_range;
  if (range[kIndirection].size() == loc.page_number) {
    spin_.lock();
    if (range[kSchemaEncoding].size() == loc.page_number) {
      range[kIndirection].push_back(std::move(Page{}));
      range[kSchemaEncoding].push_back(std::move(Page{}));
      range[kStartTime].push_back(std::move(Page{}));
      range[kPageTypeDependent].push_back(std::move(Page{}));
    }
    spin_.unlock();
  }

  auto& serializer = meta_columns_type->getSerialization();
  serializer.serialize(range[kIndirection][loc.page_number].data + loc.offset, zero_val);
  serializer.serialize(range[kSchemaEncoding][loc.page_number].data + loc.offset, zero_val);
  serializer.serialize(range[kPageTypeDependent][loc.page_number].data + loc.offset, zero_val);

  UInt64 id = static_cast<UInt64>(args.id);
  Value id_val = id;
  serializer.serialize(range[kStartTime][loc.page_number].data + loc.offset, id_val);
}

void InsertRange::insertUserData(InsertArgs& args, const Row& row) {
  PageRange& range = args.cur_range;
  auto& user_column_indexes = table_->getUserColumnIndexes();
  for (usize i = 0; i < row.size(); i += 1) {
    const Value& val = row[i];
    usize col_index = user_column_indexes[i];
    DataTypePtr type = TypeRegistry::getType(val.getType());
    if (type->haveMaximumSizeOfValue()) {
      usize type_size = type->getSizeOfValueInMemory();
      auto& serializer = type->getSerialization();
      ValueLoc loc{args.row_id, type_size};
      if (range[col_index].size() == loc.page_number) {
        spin_.lock();
        if (range[col_index].size() == loc.page_number) {
          range[col_index].push_back(std::move(Page{}));
        }
        spin_.unlock();
      }
      serializer.serialize(range[col_index][loc.page_number].data + loc.offset, val);
    } else {

    }
  }
}

} // namespace lstore
